/**
 * Perform HTTP request
 * 
 * {code:bash}
 * http www.google.com
 * {code}
 * 
 * {code:bash}
 * http uri=http://www.google.com
 * {code}
 * 
 * {code:bash}
 * http uri=http://abc.com --verbose
 * {code}
 * 
 * {code:bash}
 * http uri=http://127.0.0.1:12345/Service.cfc params="{'method':'info'}"
 * {code}
 * 
 **/	
component extends="commandbox.system.BaseCommand" excludeFromHelp=false {

	/**
	 * @uri.hint the uri to request
	 * @method.hint the HTTP method to be used
	 * @method.options GET,POST,PUT,DELETE,HEAD,TRACE,OPTIONS
	 * @charset.hint charset for the request
	 * @charset.options utf-8,iso-8859-1,windows-1252,us-ascii,shift_jis,iso-2022-jp,euc-jp,euc-kr,big5,euc-cn,utf-16
	 * @timeout.hint timeout for the request in seconds
	 * @verbose.hint if true then the whole HTTP response will be displayed, otherwise only the response content
	 * @params.hint a JSON encoded struct/object of key-value pairs to be appended to the URI
	 * @formfields.hint a JSON encoded struct/object of key-value pairs to sent as form fields in a POST request
	 * @body.hint the body to be sent (only supports string data right now)
	 * @cookies.hint a JSON encoded struct/object of key-value pairs to sent as cookies
	 * @headers.hint a JSON encoded struct/object of key-value pairs to sent as HTTP headers
	 **/
	function run(
		required string uri,
		string method="GET",
		string charset="utf-8",
		numeric timeout=60,
		boolean verbose=false,
		string params="",
		string formfields="",
		string body="",
		string cookies="",
		string headers=""
	) {

		arguments.uri = trim(arguments.uri);

		if (len(arguments.uri) eq 0) {
			return error("Parameter 'uri' cannot be an empty string");
		}

		if (!isURL(arguments.uri)) {
			return error("Parameter 'uri' needs to be a valid URI");
		}

		arguments.params = getValidatedJSONObject(name="params", value=arguments.params);
		arguments.formfields = getValidatedJSONObject(name="formfields", value=arguments.formfields);
		arguments.cookies = getValidatedJSONObject(name="cookies", value=arguments.cookies);
		arguments.headers = getValidatedJSONObject(name="headers", value=arguments.headers);

		if (hasError()) {
			return;
		}

		var key = "";

		http url="#arguments.uri#" method="#arguments.method#" timeout="#arguments.timeout#" throwOnError="false" {
			for (key in arguments.params) {
				httpparam type="url" name="#key#" value="#arguments.params[key]#";
			}
			for (key in arguments.formfields) {
				httpparam type="formfield" name="#key#" value="#arguments.formfields[key]#";
			}
			for (key in arguments.cookies) {
				httpparam type="cookie" name="#key#" value="#arguments.cookies[key]#";
			}
			for (key in arguments.headers) {
				httpparam type="header" name="#key#" value="#arguments.headers[key]#";
			}
			if (len(arguments.body) gt 0) {
				httpparam type="body" value="#arguments.body#";
			}
		}

		if (arguments.verbose) {
			print.line(serializeJSON(cfhttp));
			// TODO: only do json serialization if a specific JSON flag is set, otherwise do a friendly display of the result
		}

		if (structKeyExists(cfhttp, "fileContent") and len(cfhttp.fileContent) gt 0) {
			print.line(cfhttp.fileContent);
			// TODO: maybe do pretty print this, e.g. json indentation or maybe even html prettifying?
		}
		
		// TODO: how to do error handling on this one? (also a flag, e.g. --showerrors?)
	}

	private struct function getValidatedJSONObject(required string name, required string value) {

		if (len(arguments.value) eq 0) {
			return {};
		}
		if (!isJSON(arguments.value)) {
			error("Parameter '#arguments.name#' needs to be valid JSON.");
			return {};
		}

		var obj = deserializeJSON(arguments.value);
		if (!isStruct(obj)) {
			error("Parameter '#arguments.name#' needs to be valid JSON object.");
			return {};
		}

		for (var key in obj) {
			if (!isSimpleValue(obj[key])) {
				error("Parameter '#arguments.name#' needs to be valid JSON object with only simple values. Key '#key#' contains a complex value.");
				return {};
			}
		}

		return obj;
	}

	/**
	 * A quick way to test if a string is a URL.
	 * Regex by Gruber: http://daringfireball.net/2010/07/improved_regex_for_matching_urls
	 * 
	 * @param stringToCheck 	 The string to check. (Required)
	 * @return Returns a boolean. 
	 * @author Nathan Dintenfass (nathan@changemedia.com) 
	 * @version 2, August 5, 2010 
	 */
	private boolean function isURL(required string stringToCheck) {
		var URLRegEx = "(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'"".,<>?«»“”‘’]))";
		return isValid("regex", arguments.stringToCheck, URLRegex);
	}
}